/**
 * @file mqo.hpp
 * @date 2015-05-24
 * @brief Tv\[XR[h
 */
// for January2015v3

#pragma once

#include <Siv3D.hpp>

#define STRBUF (4096)
#define SIG L"Metasequoia Document"

/// <summary>double𕶎</summary>
String to_s(double v, int place) {
	wchar_t format[STRBUF];
	swprintf_s(format, L"%%.%df", place);
	wchar_t ws[STRBUF];
	swprintf_s(ws, format, v);

	String ret(ws);
	return ret;
}

/// <summary>Vec3𕶎</summary>
String to_s(const Vec3& v, int place) {
	wchar_t format[STRBUF];
	swprintf_s(format, L"%%.%df %%.%df %%.%df",
		place, place, place);
	wchar_t ws[STRBUF];
	swprintf_s(ws, format, v.x, v.y, v.z);

	String ret(ws);
	return ret;
}

/// <summary>Vec3𕶎</summary>
String to_s(const Vec3& v) {
	String ret = Format(v.x, L" ", v.y, L" ", v.z);
	return ret;
}

/// <summary>F𕶎</summary>
String to_s(ColorF c, int place, bool isAlpha) {
	wchar_t format[STRBUF];
	wchar_t ws[STRBUF];
	if (isAlpha == false) {
		swprintf_s(format, L"%%.%df %%.%df %%.%df",
			place, place, place);
		swprintf_s(ws, format, c.r, c.g, c.b);
	}
	else {
		swprintf_s(format, L"%%.%df %%.%df %%.%df %%.%df",
			place, place, place, place);
		swprintf_s(ws, format, c.r, c.g, c.b, c.a);
	}

	String ret(ws);
	return ret;
}

/// <summary>z𕶎</summary>
String to_s(const Array<double>& a) {
	String ret;
	for (auto v : a) {
		ret += L" " + ToString(v);
	}
	return ret.trimLeft();
}

/// <summary>z𕶎</summary>
String to_s(const Array<int>& a) {
	String ret;
	for (auto v : a) {
		ret += L" " + ToString(v);
	}
	return ret.trimLeft();
}

/// <summary>Vec2z𕶎</summary>
String to_s(const Array<Vec2>& v2s) {
	String ret;
	for (auto v2 : v2s) {
		ret += L" " + ToString(v2.x) + L" " + ToString(v2.y);
	}
	return ret.trimLeft();
}

/// <summary>wb_</summary>
struct MqoHeader {
	String sig = SIG;
	String type = L"Text";
	double ver = 1.1;
};

/// <summary>TlC</summary>
struct MqoThumbnail {
	int width = 128;
	int height = 128;
	int bitNum = 24;
	String colorFormat = L"rgb";
	String compress = L"raw";
	Array<String> lines;
};

/// <summary>Cg</summary>
struct MqoLight {
	Vec3 dir = Vec3(0.408, 0.408, 0.816);
	ColorF color = ColorF(1);
};

/// <summary>̃V[</summary>
struct MqoScene {
	Vec3 pos = Vec3(0, 0, 1500);
	Vec3 lookat = Vec3(0, 0, 0);
	double head = -0.5236;
	double pich = 0.5236;
	double bank = 0;
	int ortho = 0;
	double zoom2 = 5.0000;
	s3d::ColorF amb = ColorF(0.250, 0.250, 0.250);
	double frontclip = 225.00002;
	double backclip = 45000;
	Array<MqoLight> dirlights;
};

/// <summary>1̍ގs</summary>
struct MqoMaterial {
	/// <summary>ގ</summary>
	String name = L"mat";
	/// <summary>VF[_[</summary>
	int shader = 3;
	/// <summary>ʂǂ</summary>
	int dbls = 0;
	/// <summary>F</summary>
	s3d::ColorF col = ColorF(1);
	/// <summary>gU̖邳</summary>
	double dif = 0.8000;
	/// <summary>̖邳</summary>
	double amb = 0.600;
	/// <summary>Ȕ̖邳</summary>
	double emi = 0.000;
	double spc = 0.000;
	/// <summary>̋</summary>
	double power = 5.00;
	Optional<String> tex;
	Optional<String> aplane;
	Optional<String> bump;
};

/// <summary>ގg</summary>
struct MqoMaterialEx {
	int index;
	String shadertype;
	String shadername;
	Array<String> paramLines;
};

/// <summary>uu</summary>
struct MqoBlob {
	String type;
	double weight;
	Vec3 pos;
	Vec3 scale;
	Vec3 angle;
	double power;
};

/// <summary>1̖</summary>
struct MqoFace {
	/// <summary>_CfbNX</summary>
	Array<int> indices;
	/// <summary>ގCfbNX</summary>
	int materialIndex = 0;

	/// <summary>_UV</summary>
	Array<Vec2> uvs;
	Array<double> crss;
};

/// <summary>1̃IuWFNg</summary>
struct MqoObject {
	String name;
	int uid = -1;
	int depth = 0;
	int folding = 0;
	Vec3 scale = Vec3(1, 1, 1);
	Vec3 rotation = Vec3(0, 0, 0);
	Vec3 translation = Vec3(0, 0, 0);
	int patch = 0;
	int patchtri;
	int segment;
	int visible = 15;
	int locking = 0;
	int shading = 1;
	double facet = 59.5;
	s3d::ColorF color = ColorF(0.898, 0.498, 0.698);
	int color_type = 0;
	int mirror = 0;
	int mirror_axis = 1;
	double mirror_dis = 100;
	Array<Vec3> vertices;
	Array<MqoFace> faces;
};

/// <summary>1̃f[^</summary>
struct MqoFile {
	/// <summary>[h̃tpX</summary>
	String path;

	MqoHeader header;
	Optional<MqoThumbnail> thumbnail;
	MqoScene scene;
	//MqoBackImage backImage;
	Array<MqoMaterial> materials;
	Array<MqoMaterialEx> materialexs;
	Array<MqoObject> objects;
	Array<MqoBlob> blobs;

	/// <summary>zNA</summary>
	void clear() {
		materials.clear();
		materials.clear();
		objects.clear();
		blobs.clear();
	}
};

/// <summary>.mqot@Cǂݎ֐Q</summary>
struct MqoFiler {
	/// <summary>l3o</summary>
	Vec3 parseFloating(String line) const {
		Vec3 ret;
		auto vs = line.trim().split(L' ');
		auto n = vs.size();
		if (n >= 1) {
			ret.x = FromString<double>(vs[0]);
		}
		if (n >= 2) {
			ret.y = FromString<double>(vs[1]);
		}
		if (n >= 3) {
			ret.z = FromString<double>(vs[2]);
		}
		return ret;
	}

	/// <summary>L[Đl3o</summary>
	Vec3 parseKeyFloating(String line) const {
		Vec3 ret;
		auto vs = line.trim().split(L' ');
		auto n = vs.size();
		if (n >= 2) {
			ret.x = FromString<double>(vs[1]);
		}
		if (n >= 3) {
			ret.y = FromString<double>(vs[2]);
		}
		if (n >= 4) {
			ret.z = FromString<double>(vs[3]);
		}
		return ret;
	}

	/// <summary>_1sp[X</summary>
	Vec3 parseVertex(String line) const {
		return parseFloating(line);
	}

	/// <summary>ʂ1sp[X</summary>
	MqoFace parseFace(String line) const {
		MqoFace ret;

		std::wstring ws(line.trim().c_str());
		std::wsmatch wsm;

		int n = 0;
		std::wregex name(LR"(^(\d+))");
		if (std::regex_search(ws, wsm, name)) {
			if (wsm.size() >= 2) {
				n = FromString<int>(wsm[1].str());
			}
		}

		std::wregex v(LR"(V\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, v)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				for (auto elm : ss) {
					ret.indices.push_back(FromString<int>(elm));
				}
			}
		}

		std::wregex m(LR"(M\((\d+)\))");
		if (std::regex_search(ws, wsm, m)) {
			if (wsm.size() >= 2) {
				ret.materialIndex = FromString<int>(wsm[1].str());
			}
		}

		std::wregex uv(LR"(UV\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, uv)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				int num = ss.size();
				for (int i = 0; i < num / 2; ++i) {
					Vec2 v2(FromString<double>(ss[i * 2]), FromString<double>(ss[i * 2 + 1]));
					ret.uvs.push_back(v2);
				}
			}
		}

		/*
		std::wregex col(LR"(COL\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, col)) {
			if (wsm.size() >= 2) {
				//
			}
		}*/

		std::wregex crs(LR"(CRS\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, crs)) {
			if (wsm.size() >= 2) {
				String s(wsm[1].str());
				auto ss = s.split(L' ');
				for (auto elm : ss) {
					ret.crss.push_back(FromString<double>(elm));
				}
			}
		}

		return ret;
	}

	/// <summary>1s̍ގp[X</summary>
	MqoMaterial parseMaterial(String line) const {
		MqoMaterial ret;

		std::wstring ws(line.trim().c_str());
		std::wsmatch wsm;

		std::wregex name(LR"(^\"([^\"]*)\")");
		if (std::regex_search(ws, wsm, name)) {
			if (wsm.size() >= 2) {
				ret.name = wsm[1].str();
			}
		}

		std::wregex shader(LR"(shader\((\d+)\))");
		if (std::regex_search(ws, wsm, shader)) {
			if (wsm.size() >= 2) {
				ret.shader = FromString<int>(wsm[1].str());
			}
		}

		std::wregex dbls(LR"(dbls\((\d+)\))");
		if (std::regex_search(ws, wsm, dbls)) {
			if (wsm.size() >= 2) {
				ret.dbls = FromString<int>(wsm[1].str());
			}
		}

		std::wregex col(LR"(col\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, col)) {
			if (wsm.size() >= 2) {
				String cols = wsm[1].str();
				auto ss = cols.split(L' ');
				if (ss.size() >= 4) {
					ret.col.r = FromString<double>(ss[0]);
					ret.col.g = FromString<double>(ss[1]);
					ret.col.b = FromString<double>(ss[2]);
					ret.col.a = FromString<double>(ss[3]);
				}
			}
		}

		std::wregex dif(LR"(dif\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, dif)) {
			if (wsm.size() >= 2) {
				ret.dif = FromString<double>(wsm[1].str());
			}
		}

		std::wregex amb(LR"(amb\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, amb)) {
			if (wsm.size() >= 2) {
				ret.amb = FromString<double>(wsm[1].str());
			}
		}

		std::wregex emi(LR"(emi\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, emi)) {
			if (wsm.size() >= 2) {
				ret.emi = FromString<double>(wsm[1].str());
			}
		}

		std::wregex spc(LR"(spc\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, spc)) {
			if (wsm.size() >= 2) {
				ret.spc = FromString<double>(wsm[1].str());
			}
		}

		std::wregex power(LR"(power\(([^\)]+)\))");
		if (std::regex_search(ws, wsm, power)) {
			if (wsm.size() >= 2) {
				ret.power = FromString<double>(wsm[1].str());
			}
		}

		std::wregex tex(LR"(tex\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, tex)) {
			if (wsm.size() >= 2) {
				ret.tex.emplace(wsm[1].str());
			}
		}

		std::wregex aplane(LR"(aplane\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, aplane)) {
			if (wsm.size() >= 2) {
				ret.aplane.emplace(wsm[1].str());
			}
		}

		std::wregex bump(LR"(bump\(\"([^\"]*)\"\))");
		if (std::regex_search(ws, wsm, bump)) {
			if (wsm.size() >= 2) {
				ret.bump.emplace(wsm[1].str());
			}
		}

		return ret;
	}

	/// <summary>Scene`Nǂݎ</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool parseScene(TextReader r,MqoFile& f) const {
		bool isEof = false;
		while (!isEof) {
			Optional<String> line = r.readLine();
			if (line == false) {
				isEof = true;
				break;
			}

			String s = line.value().trim();
			if (s.includes(L"}")) {
				break;
			}

			auto ss = s.split(L' ');
			if (ss.size() < 2) {
				continue;
			}
			String key = ss[0];
			String v = ss[1];

			if (key == L"dirlights") {
				MqoLight light;
				while (!isEof) {
					line = r.readLine();
					if (line == false) {
						isEof = true;
						break;
					}

					s = line.value().trim();
					if (s.includes(L"}")) {
						f.scene.dirlights.push_back(light);
						break;
					}

					if (s.includes(L"{")) {
						while (!isEof) {
							line = r.readLine();
							if (line == false) {
								isEof = true;
								break;
							}

							s = line.value().trim();
							if (s.includes(L"}")) {
								break;
							}

							ss = s.split(L' ');
							key = ss[0];
							if (ss.size() >= 4) {
								Vec3 v3(Parse<double>(ss[1]),
									Parse<double>(ss[2]),
									Parse<double>(ss[3]));
								if (key == L"dir") {
									light.dir = v3;
								}
								else if (key == L"color") {
									light.color = ColorF(v3);
								}
							}

						}
					}

				} // dirlights

			}
			else if (key == L"head") {
				f.scene.head = Parse<double>(v);
			}
			else if (key == L"ortho") {
				f.scene.ortho = Parse<int>(v);
			}
			else if (key == L"zoom2") {
				f.scene.zoom2 = Parse<double>(v);
			}
			else if (key == L"amb") {
				if (ss.size() >= 4) {
					f.scene.amb = ColorF(Parse<double>(ss[1]),
						Parse<double>(ss[2]),
						Parse<double>(ss[3]));
				}
			}
		}

		return isEof;
	}

	/// <summary>Object`Nǂݎ</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool parseObject(TextReader r,MqoObject& obj) const {
		bool isEof = false;

		while (!isEof) {
			auto line = r.readLine();
			if (line == false) {
				isEof = true;
				break;
			}

			String s = line.value().trim();
			if (s.includes(L"}")) {
				break;
			}

			if (s.includes(L"BVertex ")) {
				isEof = true;
				break;
			}

			auto ss = s.split(L' ');
			int num = ss.size();
			String key = ss[0];

			if (key == L"depth") {
				if (num >= 2) {
					obj.depth = Parse<int>(ss[1]);
				}
			}
			else if (key == L"folding") {
				if (num >= 2) {
					obj.folding = Parse<int>(ss[1]);
				}
			}
			else if (key == L"scale") {
				if (num >= 4) {
					obj.scale = parseKeyFloating(s);
				}
			}
			else if (key == L"rotation") {
				if (num >= 4) {
					obj.rotation = parseKeyFloating(s);
				}
			}
			else if (key == L"translation") {
				if (num >= 4) {
					obj.translation = parseKeyFloating(s);
				}
			}
			else if (key == L"visible") {
				if (num >= 2) {
					obj.visible = Parse<int>(ss[1]);
				}
			}
			else if (key == L"locking") {
				if (num >= 2) {
					obj.locking = Parse<int>(ss[1]);
				}
			}
			else if (key == L"shading") {
				if (num >= 2) {
					obj.shading = Parse<int>(ss[1]);
				}
			}
			else if (key == L"facet") {
				if (num >= 2) {
					obj.facet = Parse<double>(ss[1]);
				}
			}
			else if (key == L"color") {
				if (num >= 4) {
					obj.color = ColorF(parseKeyFloating(s));
				}
			}
			else if (key == L"color_type") {
				if (num >= 2) {
					obj.color_type = Parse<int>(ss[1]);
				}
			}
			else if (key == L"vertex") {
				while (!isEof) {
					line = r.readLine();
					if (line == false) {
						isEof = true;
						break;
					}
					s = line.value().trim();
					if (s.includes(L"}")) {
						break;
					}

					auto v = this->parseVertex(s);
					obj.vertices.push_back(v);
				}
			}
			else if (key == L"vertexattr") {
				isEof = skipUnknown(r);
			}
			else if (key == L"face") {
				while (!isEof) {
					line = r.readLine();
					if (line == false) {
						isEof = true;
						break;
					}
					s = line.value().trim();
					if (s.includes(L"}")) {
						break;
					}

					auto face = this->parseFace(s);
					obj.faces.push_back(face);
				}
			}
			else if (s.includes(L"{")) {
				isEof = skipUnknown(r);
			}

		}
		return isEof;
	}

	/// <summary>mȂ`Nǂݔ΂</summary>
	/// <returns>isEof̏ԂԂ</returns>
	bool skipUnknown(TextReader r) const {
		bool isEof = false;
		int layer = 1;
		while (!isEof) {
			auto line = r.readLine();
			if (line == false) {
				isEof = true;
				break;
			}
			auto s = line.value();
			if (s.includes(L"{")) {
				++layer;
			}
			else if (s.includes(L"}")) {
				--layer;
				if (layer <= 0) {
					break;
				}
			}
		}
		return isEof;
	}

	/// <summary>.mqot@Cǂݎ</summary>
	/// <param name="r">Shift_JIS̃eLXg[_</param>
	/// <param name="f">ʂmۂ</param>
	bool LoadMqo(TextReader r,MqoFile& f) const {
		try {
			Optional<String> line;

			bool isEof = false;

			line = r.readLine();
			if (line == false) {
				return false;
			}
			if (line.value() != SIG) {
				return false;
			}

			line = r.readLine();

			while (!isEof) {
				line = r.readLine();
				if (line == false) {
					isEof = true;
					break;
				}
				String s = line.value();
				if (s.startsWith(L"Eof")) {
					isEof = true;
					break;
				}

				if (s.startsWith(L"Thumbnail")) {
					MqoThumbnail tn;
					while (!isEof) {
						line = r.readLine();
						if (line == false) {
							isEof = true;
							break;
						}

						s = line.value();
						if (s.includes(L"}")) {
							f.thumbnail = tn;
							break;
						}

						tn.lines.push_back(s);
					}
				}
				else if (s.startsWith(L"Scene")) {
					isEof = parseScene(r,f);
				}
				else if (s.startsWith(L"BackImage")) {
					isEof = skipUnknown(r);
				}
				else if (s.startsWith(L"Material")) {
					while (!isEof) {
						line = r.readLine();
						if (line == false) {
							isEof = true;
							break;
						}

						s = line.value().trim();
						if (s.includes(L"}")) {
							break;
						}

						auto m = this->parseMaterial(s);
						f.materials.push_back(m);
					}
				}
				else if (s.startsWith(L"Object")) {
					MqoObject obj;
					auto ss = s.split(L'\"');
					if (ss.size() >= 3) {
						obj.name = ss[1];
					}

					isEof = parseObject(r, obj);
					if (isEof == false) {
						f.objects.push_back(obj);
					}
				}
				else if (s.startsWith(L"Blob")) {
					isEof = skipUnknown(r);
				}
				else if (s.includes(L"{")) {
					isEof = skipUnknown(r);
				}
			}
		}
		catch (...) {
			return false;
		}
		return true;
	}

	/// <summary>B
	/// .mqot@Co</summary>
	/// <param name="f">oΏۂ̃t@C</param>
	/// <param name="w">Shift_JIS̃eLXgC^[</param>
	void SaveMqo(const MqoFile& f, TextWriter w) const {
		{ // wb_
			w.writeln(f.header.sig);
			w.writeln(L"Format ", f.header.type, L" Ver ", ToString(f.header.ver, 1));
		}
		w.writeln<String>(L"");

		if (f.thumbnail) { // Tl
			const auto& tn = f.thumbnail.value();
			w.writeln(L"Thumbnail ",
				tn.width, L" ", tn.height,
				L" ", tn.bitNum,
				L" ", tn.colorFormat,
				L" ", tn.compress, L" {");
			for (const auto& s : tn.lines) {
				w.writeln(L"\t", s);
			}
			w.writeln(L"}");
		}

		w.writeln(L"Scene {");
		w.writeln(L"\tpos ", to_s(f.scene.pos, 4));
		w.writeln(L"\tlookat ", to_s(f.scene.lookat, 4));
		w.writeln(L"\thead ", to_s(f.scene.head, 4));
		w.writeln(L"\tpich ", to_s(f.scene.pich, 4));
		w.writeln(L"\tbank ", to_s(f.scene.bank, 4));
		w.writeln(L"\tortho ", f.scene.ortho);
		w.writeln(L"\tzoom2 ", to_s(f.scene.zoom2, 4));
		w.writeln(L"\tfrontclip ", to_s(f.scene.frontclip,5));
		w.writeln(L"\tbackclip ", f.scene.backclip);
		w.writeln(L"\tamb ", to_s(f.scene.amb, 3, false));
		int lightNum = f.scene.dirlights.size();
		w.writeln(L"\tdirlights ", lightNum, L" {");
		for (int i = 0; i < lightNum; ++i) {
			MqoLight light = f.scene.dirlights[i];
			w.writeln(L"\t\tlight {");
			w.writeln(L"\t\t\tdir ", to_s(light.dir, 3));
			w.writeln(L"\t\t\tcolor ", to_s(light.color, 3, false));
			w.writeln(L"\t\t}");
		}
		w.writeln(L"\t}");
		w.writeln(L"}");

		int n = f.materials.size();
		w.writeln(L"Material ", n, L" {");
		for (int i = 0; i < n; ++i) {
			MqoMaterial mat = f.materials[i];
			w.write(L"\t\"", mat.name, L"\"");
			w.write(L" shader(", mat.shader, L")");
			if (mat.dbls > 0) {
				w.write(L" dbls(", mat.dbls, L")");
			}
			w.write(L" col(", to_s(mat.col, 3, false), L")");
			w.write(L" dif(", to_s(mat.dif, 3), L")");
			w.write(L" amb(", to_s(mat.amb, 3), L")");
			w.write(L" emi(", to_s(mat.emi, 3), L")");
			w.write(L" spc(", to_s(mat.spc, 3), L")");
			w.write(L" power(", to_s(mat.power, 2), L")");
			if (mat.tex) {
				w.write(L" tex(\"", mat.tex.value(), L"\")");
			}
			if (mat.aplane) {
				w.write(L" aplane(\"", mat.aplane.value(), L"\")");
			}
			if (mat.bump) {
				w.write(L" bump(\"", mat.bump.value(), L"\")");
			}
			w.writeln(L"");
		}
		w.writeln(L"}");

		n = f.materialexs.size();
		if (n > 0) {
			w.writeln(L"MaterialEx ", n, L" {");
			for (const auto& mat : f.materialexs) {
				w.writeln(L"\tmaterial ", mat.index, L" {");
				w.writeln(L"\t\t shadertype \"", mat.shadertype, L"\"");
				w.writeln(L"\t\t shadername \"", mat.shadername, L"\"");
				w.writeln(L"\t\t shaderparam 2 {");
				w.writeln(L"\t\t\tToon int 1"); // 
				w.writeln(L"\t\t\tEdge bool 1"); // 
				w.writeln(L"\t\t}");
				w.writeln(L"\t}");
			}
			w.writeln(L"}");
		}

		n = f.objects.size();
		for (int i = 0; i < n; ++i) { // obj
			MqoObject obj = f.objects[i];
			w.writeln(L"Object \"", obj.name, L"\" {");
			w.writeln(L"\tdepth ", obj.depth);
			w.writeln(L"\tfolding ", obj.folding);
			w.writeln(L"\tscale ", to_s(obj.scale, 6));
			w.writeln(L"\trotation ", to_s(obj.rotation, 6));
			w.writeln(L"\ttranslation ", to_s(obj.translation, 6));
			if (obj.patch > 0) {
				w.writeln(L"\tpatch ", obj.patch);
				w.writeln(L"\tpatchtri ", obj.patchtri);
			}
			w.writeln(L"\tvisible ", obj.visible);
			w.writeln(L"\tlocking ", obj.locking);
			w.writeln(L"\tshading ", obj.shading);
			w.writeln(L"\tfacet ", to_s(obj.facet, 1));
			w.writeln(L"\tcolor ", to_s(obj.color, 3, false));
			w.writeln(L"\tcolor_type ", obj.color_type);
			if (obj.mirror > 0) {
				w.writeln(L"\tmirror ", obj.mirror);
				w.writeln(L"\tmirror_axis ", obj.mirror_axis);
				w.writeln(L"\tmirror_dis ", to_s(obj.mirror_dis, 3));
			}

			int vertexNum = obj.vertices.size();
			w.writeln(L"\tvertex ", vertexNum, L" {");
			for (int j = 0; j < vertexNum; ++j) { // vertex
				Vec3 vtx = obj.vertices[j];
				w.writeln(L"\t\t", to_s(vtx,4));
			}
			w.writeln<String>(L"\t}");

			int attrNum = 0;
			for (int j = 0; j < attrNum; ++j) {
				// ĂȂ
			}

			int faceNum = obj.faces.size();
			w.writeln(L"\tface ", faceNum, L" {");
			for (int j = 0; j < faceNum; ++j) { // face
				MqoFace face = obj.faces[j];
				int vNum = face.indices.size();
				w.write(L"\t\t", vNum, L" V(", to_s(face.indices), L")");
				w.write(L" M(", face.materialIndex, L")");
				w.write(L" UV(", to_s(face.uvs), L")");
				if (face.crss.size() > 0) {
					w.write(L" CRS(", to_s(face.crss), L")");
				}
				w.writeln(L"");
			}
			w.writeln(L"\t}");
		} // obj

		int blobNum = f.blobs.size();
		if (blobNum > 0) {
			w.writeln(L"Blob blob1 ", blobNum, L" {");
			for (const auto& blob: f.blobs) {
				w.writeln(L"\t", blob.type, L" {");
				w.writeln(L"\t\tpos ",to_s(blob.pos,4));
				w.writeln(L"\t\tscale ",to_s(blob.scale,4));
				w.writeln(L"\t\tangle ",to_s(blob.angle,4));
				w.writeln(L"\t\tpower ",to_s(blob.power,4));
				w.writeln(L"\t}");
			}
			w.writeln(L"}");
		}

		w.writeln(L"Eof");
	}
};
